// Schaltschrankuhr "Chronoverter"

// Standard-IO fr AVRs:
#include <avr/io.h>

// Bits wie Variablen benutzen:
//#include "sbit.h"

// dann sind auch Interrupts notwendig:
#include <avr/interrupt.h>

#define NOP asm volatile ("nop")

// #define DEBUG    // Wenn definiert luft das Ganze 60 mal schneller (1 Minute entspricht 1 Sekunde)

// Pin Definitionen
// Port A
#define IN0     0
#define IN1     1
#define IN2     2
#define IN3     3
#define IN4     4
#define IN5     5
#define OSCOUT  6
#define OSCIN   7

// Port B
#define LAMP   0
#define PWMA   1  // Minuten
#define BEEP   2
#define PWMB   3  // Stunden
#define LED    4
#define CLOCK  6

#define BTN_SETH     _BV(IN0) // PA0
#define BTN_SETM     _BV(IN1) // PA1
#define BTN_STOP     _BV(IN2) // PA2
#define BTN_ALMODE   _BV(IN3) // PA3
#define BTN_CNTMODE  _BV(IN4) // PA4
#define BTN_ENALARM  _BV(IN5) // PA5

// Bits in state variable
#define ALARMENABLE   1   // Alarm!
#define HZ2           2   // 2Hz Takt, wird fr Blinken und periodisches Piepsen genutzt

#define ALARMDAUER  255   // Alarmdauer in Minuten, 255 = unendlich
#define PIEPFREQUENZ 20   // Frequenz des Piepsers f=33.333kHz / 2*PIEPFREQUENZ

#define ALARMOFF {PORTB |= _BV(LAMP); alarm_enable&=~_BV(ALARMENABLE); /* TIMSK &= ~_BV(TOIE1);*/}

volatile int zeit=-1;     // normale Zeit in Minuten seit 0:00 Uhr, -1 = nicht gestellt
volatile int weckzeit=-1; // Weckzeit in Minuten seit 0:00 Uhr, -1 = nicht gestellt
volatile int countdown=-1;// Countown in Minuten
volatile int *mode=&zeit; // zeigt auf eine der drei mglichen Zeitzhler (s.o.) und gibt an was gerade angezeigt wird
volatile int timeout=0;   // Timeout fr Stromausfallerkennung
volatile char alarm=0;    // wenn >0 dann blinkt und piepst es
volatile char countdownstart=0; // wenn >0 dann luft der Countdown (so er denn gestellt ist)
volatile char alarm_enable=0; // kein Alarm wenn entsprechender Schalter aus

// Hardware initialisieren
void init(void)
{
	PORTA = 0;
	DDRA  = _BV(OSCOUT);
	PORTB = _BV(LAMP);
	DDRB  = _BV(LAMP) + _BV(PWMA) + _BV(BEEP) + _BV(PWMB) + _BV(LED);
	// init Timer 0 (100 Hz Timer)
	TIMSK = _BV(TOIE0);
	TCCR0 = _BV(CS02) + _BV(CS01) + _BV(CS00); // enable external clock fr Timer0 (100Hz Netzfrequenz)
	TCNT0 = 256-50;
	// init Timer 1 (PWM)
	TCCR1A = _BV(PWM1A) + _BV(PWM1B) + _BV(COM1A1) + _BV(COM1B1);
	TCCR1B = _BV(CTC1) + _BV(CS10);
	OCR1C  = 240;
	OCR1A  = 0;
	OCR1B  = 0;
	TIMSK |= _BV(TOIE1);
	sei();
}

// setze Anzeigemodus und aktualisiere Anzeige bei Moduswechsel
// wenn m==NULL wird nur die Anzeige aktualisiert
void setmode(volatile int *m)
{
	int t;
	
	// wenn Mode sich nicht verndert hat
	if ((m) && (mode!=m))
	{
		mode=m;
		m=0;
	}	
	if (!m)
	{
		t=*mode;
		if (t<0) t=0;
		OCR1A=((t%60)*4);      // Minuten in 1/4 Minuten ausgeben
		OCR1B=((t/60)*10);     // Stunden ausgeben
	}
}			

// Timer1 ISR 33kHz fr PWM und piepsen
ISR(TIMER1_OVF1_vect)
{
	static int piep=0;
	static int usec;

	// Test ob lnger als 10ms kein 100Hz Takt -> Stromausfall
	if (timeout>17000)
	{
		// Timer auf Batteriebetrieb stellen
		TCCR1A = 0;
		TCCR1B = _BV(CTC1) + _BV(CS11) + _BV(CS12) + _BV(CS13);
		OCR1C  = 244;
		DDRB=0;
		PORTB=0;
		usec++;
		if (usec>241)
		{
			zeit++;
			if (zeit>1439) zeit=0;
			usec=1;
		}
	}
	else // wenn 100Hz Takt vorhanden...
	{
		
		timeout++;
		// ...und wenn im Batteriemodus...
		if (usec)
		{
			usec=0;
			// ...Timer auf Normalbetrieb stellen
			init();
			setmode(0);
		}
		
		// Mit LED blinken im Halbsekundentakt
		if (alarm_enable & _BV(HZ2))
			PORTB |= _BV(LED);
		else
			PORTB &= ~_BV(LED);
		
		// Alarme behandeln
		if (alarm_enable & _BV(HZ2))  // wenn "gerade" halbe Sekunde...
		{
			if (alarm>0)          // ...und wenn Alarm...
			{
				piep++;	      // ...dann piepsen
				if (piep>PIEPFREQUENZ)
				{
					piep=0;
					PORTB ^= _BV(BEEP);
				}
			}
			if ((alarm>0) || (zeit<0)) // wenn Alarm oder Zeit nicht gestellt
				PORTB &= ~_BV(LAMP);  // Lampe an
		}
		else
			PORTB |= _BV(LAMP);
	}

}

// Timer0 ISR wird zweimal pro Sekunde aufgerufen
ISR(TIMER0_OVF0_vect)
{
	static char sec;
	
	TCNT0 = 256-50;
	timeout = 0;
	// Blinke mit Lampe und LED im 1/2 Sekunden Takt
	alarm_enable ^= _BV(HZ2);
	sec++;
	if (sec>=120) // 120 halbe Sekunden sind eine Minute
	{
		sec=0;

#ifdef DEBUG
	}
#endif
		// Alarmdauer runterzhlen
		if ((alarm>0) && (alarm!=255))
		{
			alarm--;
			if (alarm==0) ALARMOFF;
		}
		
		// Zeit hochzhlen in 1/4 Sekunden seit 0:00 Uhr
		if (zeit>=0) // nur hochzhlen wenn bereits gestellt
		{
			zeit++;
			if (zeit>1439) zeit=0; // der Tag ist um, wieder bei null anfangen
			if ((weckzeit==zeit) && !(PINA & BTN_ENALARM)) alarm=ALARMDAUER;
		}
		
		// Countdown runterzhlen
		if ((countdown>0) && (countdownstart>0))
		{
			countdown--;
			if (countdown==0)
			{
				alarm=ALARMDAUER;
				countdownstart=0;
			}
		}
		
		// Anzeige aktualisieren		
		setmode(0);
  	
#ifndef DEBUG
	}
#endif
}

char getkeys(void)
{
	char cnt=0, oldkey=0, key;
	// zum Entprellen darauf warten, da diverse male hintereinander der selbe Wert gelesen wird
	do
	{
		key=~(PINA & 0x3f);
		NOP; NOP; NOP;
		NOP; NOP; NOP;
		NOP; NOP; NOP;
		if (key==oldkey) 
			cnt++;
		else
		{
			oldkey=key;
			cnt=0;
		}
	} while (cnt<150);
	return key;
}


int main(void)
{
	char keys;
	
	init();
	while(1)
	{
		// Taster abfragen
		keys=getkeys();

		// Displaymode Schalter abfragen		
		if (keys & BTN_ALMODE)       // wenn Weckzeitschalter an...
			setmode(&weckzeit);  // ..Weckzeit anzeigen/stellen
		else if (keys & BTN_CNTMODE) // wenn Timerschalter an...
			setmode(&countdown); // ..Timer anzeigen/stellen (Timerstart nur in diesem Modus)
		else                         // Wenn keiner der Schalter an...
		  setmode(&zeit);            // ..Zeit anzeigen/stellen

		// Stelltaster abfragen
		// Mode ist ein Pointer auf eine von drei verstellbaren Variablen:
		// countdown : Countdownzhler fr den Timer
		// weckzeit  : Was das wohl sein mag
		// zeit      : fast genauso rtselhaft
		// nun braucht nur noch der _Wert_ auf den der Pointer zeigt verstellt zu werden
		// und schon stellt man Timer, Wecker oder Zeit
		if (keys & BTN_SETH)
		{
			if (mode==&countdown) countdownstart=0; // Evtl. laufender Timer wird gestoppt
			if (*mode<0)  // negative Werte heissen, das noch nichts eingestellt war
				*mode=0;
			else
				*mode=(*mode+60)%1440;
			setmode(0); // Anzeige aktualisieren
			while (getkeys() & BTN_SETH);
			ALARMOFF;
		}
		if (keys & BTN_SETM)
		{
			if (mode==&countdown) countdownstart=0; // Evtl. laufender Timer wird gestoppt
			if (*mode<0)
				*mode=0; // negative Werte heissen, das noch nichts eingestellt war
			else
				*mode=(*mode+1)%1440;
			setmode(0); // Anzeige aktualisieren
			while (getkeys() & BTN_SETM);
			ALARMOFF;
		}
		if (keys & BTN_STOP)
		{
			if (alarm>0)  // wenn Alarm ansteht...
			{
				alarm=0;  // ..Alarm ausmachen
				ALARMOFF;
			}
			if (countdown>0) countdownstart=1; // wenn ein Timer gestellt ist Timer starten
		}
	}
}
